"use strict";
var __createBinding = (this && this.__createBinding) || (Object.create ? (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    Object.defineProperty(o, k2, { enumerable: true, get: function() { return m[k]; } });
}) : (function(o, m, k, k2) {
    if (k2 === undefined) k2 = k;
    o[k2] = m[k];
}));
var __setModuleDefault = (this && this.__setModuleDefault) || (Object.create ? (function(o, v) {
    Object.defineProperty(o, "default", { enumerable: true, value: v });
}) : function(o, v) {
    o["default"] = v;
});
var __importStar = (this && this.__importStar) || function (mod) {
    if (mod && mod.__esModule) return mod;
    var result = {};
    if (mod != null) for (var k in mod) if (k !== "default" && Object.prototype.hasOwnProperty.call(mod, k)) __createBinding(result, mod, k);
    __setModuleDefault(result, mod);
    return result;
};
var __importDefault = (this && this.__importDefault) || function (mod) {
    return (mod && mod.__esModule) ? mod : { "default": mod };
};
Object.defineProperty(exports, "__esModule", { value: true });
exports.InvoiceServer = void 0;
const bjs = __importStar(require("bitcoinjs-lib"));
const bip32 = __importStar(require("bip32"));
const level_1 = __importDefault(require("level"));
const uuid_1 = require("uuid");
const events_1 = require("events");
const networks_1 = require("./utils/networks");
const invoice_1 = require("./invoice");
class InvoiceServer extends events_1.EventEmitter {
    constructor(xpub, network, blockchainClient, checkInterval = 10000) {
        super();
        this.intervals = {};
        this.network = networks_1.Networks[network]; // TODO: Bech32only
        this.xpub = xpub;
        this.baseNode = bip32.fromBase58(xpub, this.network);
        this.blockchainClient = blockchainClient;
        this.checkInterval = checkInterval;
        this.invoiceDB = level_1.default("db-invoices", { valueEncoding: "json" });
        this.historyDB = level_1.default("db-history", { valueEncoding: "json" });
        this.addressesDB = level_1.default("db-addresses", { valueEncoding: "json" });
    }
    async start() {
        try {
            await this.addressesDB.get("used");
        }
        catch (e) {
            if (e.type === "NotFoundError") {
                await this.addressesDB.put("used", []);
            }
            else
                throw e;
        }
        return new Promise((resolve) => {
            this.invoiceDB
                .createValueStream()
                .on("data", (invoice) => {
                this._watchInvoice(invoice);
            })
                .on("end", () => {
                resolve();
            });
        });
    }
    async newInvoice(options) {
        const defaults = {
            secs: 3600,
            requiredConfirmations: 1,
        };
        options = Object.assign({}, defaults, options);
        const id = uuid_1.v4();
        const address = await this._getNewAddress();
        const expiration = Date.now() + options.secs * 1000;
        const invoice = {
            id,
            address,
            amount: options.amount,
            expiration,
            requiredConfirmations: options.requiredConfirmations,
            data: options.data,
            status: invoice_1.INVOICE_STATUS.ACTIVE,
        };
        await this._addInvoice(invoice);
        await this._watchInvoice(invoice);
        return invoice;
    }
    async _isAddressInUse(address) {
        const usedAddresses = await this.addressesDB.get("used");
        return usedAddresses.includes(address);
    }
    async _getNewAddress() {
        // TODO: choose address type
        for (let i = 0;; i++) {
            const publicKey = this.baseNode.derive(0).derive(i).publicKey;
            const address = bjs.payments.p2wpkh({
                pubkey: publicKey,
                network: this.network,
            }).address;
            if (await this._isAddressInUse(address))
                continue;
            const addressTransactions = await this.blockchainClient.getAddressTransactions(address);
            if (addressTransactions.length === 0) {
                return address;
            }
        }
    }
    async _saveInvoice(invoice) {
        return this.invoiceDB.put(invoice.id, invoice);
    }
    async _archiveInvoice(invoice) {
        if (invoice.id in this.intervals) {
            clearInterval(this.intervals[invoice.id]);
            delete this.intervals[invoice.id];
        }
        if (invoice.status === invoice_1.INVOICE_STATUS.EXPIRED) {
            this._removeUsedAddress(invoice.address);
        }
        await this.historyDB.put(invoice.id, invoice);
        await this.invoiceDB.del(invoice.id);
    }
    async _checkInvoice(id) {
        const invoice = await this.invoiceDB.get(id);
        const txs = await this.blockchainClient.getAddressTransactions(invoice.address);
        const tx = txs.find((_tx) => _tx.amount >= invoice.amount);
        if (tx) {
            if (tx.confirmations >= invoice.requiredConfirmations) {
                invoice.tx = tx.hash;
                invoice.status = invoice_1.INVOICE_STATUS.PAYMENT_CONFIRMED;
                this.emit("invoice_update", invoice);
                await this._archiveInvoice(invoice);
                return true;
            }
            else if (invoice.tx !== tx.hash) {
                invoice.tx = tx.hash;
                invoice.status = invoice_1.INVOICE_STATUS.PAYMENT_RECEIVED;
                this.emit("invoice_update", invoice);
                await this._saveInvoice(invoice);
                return true;
            }
        }
        if (Date.now() > invoice.expiration) {
            invoice.status = invoice_1.INVOICE_STATUS.EXPIRED;
            this.emit("invoice_update", invoice);
            await this._archiveInvoice(invoice);
            return true;
        }
        return false;
    }
    async _addUsedAddress(address) {
        const usedAddresses = await this.addressesDB.get("used");
        usedAddresses.push(address);
        await this.addressesDB.put("used", usedAddresses);
    }
    async _removeUsedAddress(address) {
        const usedAddresses = await this.addressesDB.get("used");
        usedAddresses.splice(usedAddresses.indexOf(address), 1);
        await this.addressesDB.put("used", usedAddresses);
    }
    async _addInvoice(invoice) {
        await this._addUsedAddress(invoice.address);
        await this.invoiceDB.put(invoice.id, invoice);
    }
    async _watchInvoice(invoice) {
        const resolved = await this._checkInvoice(invoice.id);
        if (resolved)
            return;
        this.intervals[invoice.id] = setInterval(() => {
            this._checkInvoice(invoice.id);
        }, this.checkInterval);
    }
}
exports.InvoiceServer = InvoiceServer;
